@@ -2,18 +2,24 @@ from __future__ import annotations
 
 import enum
 from collections import OrderedDict
+from collections.abc import Iterable, Mapping
 from datetime import timezone
 from itertools import starmap
-from typing import TYPE_CHECKING, Iterable, Mapping
+from typing import TYPE_CHECKING
 
-from narwhals.utils import _DeferredIterable, isinstance_or_issubclass
+from narwhals._utils import (
+    _DeferredIterable,
+    isinstance_or_issubclass,
+    qualified_type_name,
+)
 
 if TYPE_CHECKING:
-    from typing import Iterator, Sequence
+    from collections.abc import Iterator, Sequence
+    from typing import Any
 
-    from typing_extensions import Self
+    from typing_extensions import Self, TypeIs
 
-    from narwhals.typing import TimeUnit
+    from narwhals.typing import IntoDType, TimeUnit
 
 
 def _validate_dtype(dtype: DType | type[DType]) -> None:
@@ -25,44 +31,117 @@ def _validate_dtype(dtype: DType | type[DType]) -> None:
         raise TypeError(msg)
 
 
+def _is_into_dtype(obj: Any) -> TypeIs[IntoDType]:
+    return isinstance(obj, DType) or (
+        isinstance(obj, type)
+        and issubclass(obj, DType)
+        and not issubclass(obj, NestedType)
+    )
+
+
+def _is_nested_type(obj: Any) -> TypeIs[type[NestedType]]:
+    return isinstance(obj, type) and issubclass(obj, NestedType)
+
+
+def _validate_into_dtype(dtype: Any) -> None:
+    if not _is_into_dtype(dtype):
+        if _is_nested_type(dtype):
+            name = f"nw.{dtype.__name__}"
+            msg = (
+                f"{name!r} is not valid in this context.\n\n"
+                f"Hint: instead of:\n\n"
+                f"    {name}\n\n"
+                "use:\n\n"
+                f"    {name}(...)"
+            )
+        else:
+            msg = f"Expected Narwhals dtype, got: {qualified_type_name(dtype)!r}."
+        raise TypeError(msg)
+
+
 class DType:
     def __repr__(self) -> str:  # pragma: no cover
         return self.__class__.__qualname__
 
+    @classmethod
+    def base_type(cls) -> type[Self]:
+        """Return this DType's fundamental/root type class.
+
+        Examples:
+            >>> import narwhals as nw
+            >>> nw.Datetime("us").base_type()
+            <class 'narwhals.dtypes.Datetime'>
+
+            >>> nw.String.base_type()
+            <class 'narwhals.dtypes.String'>
+
+            >>> nw.List(nw.Int64).base_type()
+            <class 'narwhals.dtypes.List'>
+        """
+        return cls
+
     @classmethod
     def is_numeric(cls: type[Self]) -> bool:
+        """Check whether the data type is a numeric type."""
         return issubclass(cls, NumericType)
 
     @classmethod
     def is_integer(cls: type[Self]) -> bool:
+        """Check whether the data type is an integer type."""
         return issubclass(cls, IntegerType)
 
     @classmethod
     def is_signed_integer(cls: type[Self]) -> bool:
+        """Check whether the data type is a signed integer type."""
         return issubclass(cls, SignedIntegerType)
 
     @classmethod
     def is_unsigned_integer(cls: type[Self]) -> bool:
+        """Check whether the data type is an unsigned integer type."""
         return issubclass(cls, UnsignedIntegerType)
 
     @classmethod
     def is_float(cls: type[Self]) -> bool:
+        """Check whether the data type is a floating point type."""
         return issubclass(cls, FloatType)
 
     @classmethod
     def is_decimal(cls: type[Self]) -> bool:
+        """Check whether the data type is a decimal type."""
         return issubclass(cls, Decimal)
 
     @classmethod
     def is_temporal(cls: type[Self]) -> bool:
+        """Check whether the data type is a temporal type."""
         return issubclass(cls, TemporalType)
 
     @classmethod
     def is_nested(cls: type[Self]) -> bool:
+        """Check whether the data type is a nested type."""
         return issubclass(cls, NestedType)
 
+    @classmethod
+    def is_boolean(cls: type[Self]) -> bool:
+        """Check whether the data type is a boolean type."""
+        return issubclass(cls, Boolean)
+
     def __eq__(self, other: DType | type[DType]) -> bool:  # type: ignore[override]
-        from narwhals.utils import isinstance_or_issubclass
+        """Check if this DType is equivalent to another DType.
+
+        Examples:
+            >>> import narwhals as nw
+            >>> nw.String() == nw.String()
+            True
+            >>> nw.String() == nw.String
+            True
+            >>> nw.Int16() == nw.Int32
+            False
+            >>> nw.Boolean() == nw.Int8
+            False
+            >>> nw.Date() == nw.Datetime
+            False
+        """
+        from narwhals._utils import isinstance_or_issubclass
 
         return isinstance_or_issubclass(other, type(self))
 
@@ -125,7 +204,7 @@ class Int128(SignedIntegerType):
 
         >>> s.cast(nw.Int128).dtype
         Int128
-        >>> nw.from_native(rel).schema["a"]
+        >>> nw.from_native(rel).collect_schema()["a"]
         Int128
     """
 
@@ -191,7 +270,7 @@ class UInt128(UnsignedIntegerType):
         >>> import duckdb
         >>> df_native = pd.DataFrame({"a": [2, 1, 3, 7]})
         >>> rel = duckdb.sql(" SELECT CAST (a AS UINT128) AS a FROM df_native ")
-        >>> nw.from_native(rel).schema["a"]
+        >>> nw.from_native(rel).collect_schema()["a"]
         UInt128
     """
 
@@ -326,10 +405,12 @@ class Unknown(DType):
 class _DatetimeMeta(type):
     @property
     def time_unit(cls) -> TimeUnit:
+        """Unit of time. Defaults to `'us'` (microseconds)."""
         return "us"
 
     @property
     def time_zone(cls) -> str | None:
+        """Time zone string. Defaults to `None`."""
         return None
 
 
@@ -371,16 +452,39 @@ class Datetime(TemporalType, metaclass=_DatetimeMeta):
             time_zone = str(time_zone)
 
         self.time_unit: TimeUnit = time_unit
+        """Unit of time."""
         self.time_zone: str | None = time_zone
+        """Time zone string, as defined in zoneinfo.
 
-    def __eq__(self, other: object) -> bool:
-        # allow comparing object instances to class
+        Notes:
+            To see valid strings run `import zoneinfo; zoneinfo.available_timezones()` for a full list.
+        """
+
+    def __eq__(self, other: DType | type[DType]) -> bool:  # type: ignore[override]
+        """Check if this Datetime is equivalent to another DType.
+
+        Examples:
+            >>> import narwhals as nw
+            >>> nw.Datetime("s") == nw.Datetime("s")
+            True
+            >>> nw.Datetime() == nw.Datetime("us")
+            True
+            >>> nw.Datetime("us") == nw.Datetime("ns")
+            False
+            >>> nw.Datetime("us", "UTC") == nw.Datetime(time_unit="us", time_zone="UTC")
+            True
+            >>> nw.Datetime(time_zone="UTC") == nw.Datetime(time_zone="EST")
+            False
+            >>> nw.Datetime() == nw.Duration()
+            False
+            >>> nw.Datetime("ms") == nw.Datetime
+            True
+        """
         if type(other) is _DatetimeMeta:
             return True
-        elif isinstance(other, self.__class__):
+        if isinstance(other, self.__class__):
             return self.time_unit == other.time_unit and self.time_zone == other.time_zone
-        else:  # pragma: no cover
-            return False
+        return False  # pragma: no cover
 
     def __hash__(self) -> int:  # pragma: no cover
         return hash((self.__class__, self.time_unit, self.time_zone))
@@ -393,6 +497,7 @@ class Datetime(TemporalType, metaclass=_DatetimeMeta):
 class _DurationMeta(type):
     @property
     def time_unit(cls) -> TimeUnit:
+        """Unit of time. Defaults to `'us'` (microseconds)."""
         return "us"
 
 
@@ -425,15 +530,29 @@ class Duration(TemporalType, metaclass=_DurationMeta):
             raise ValueError(msg)
 
         self.time_unit: TimeUnit = time_unit
+        """Unit of time."""
 
-    def __eq__(self, other: object) -> bool:
-        # allow comparing object instances to class
+    def __eq__(self, other: DType | type[DType]) -> bool:  # type: ignore[override]
+        """Check if this Duration is equivalent to another DType.
+
+        Examples:
+            >>> import narwhals as nw
+            >>> nw.Duration("us") == nw.Duration("us")
+            True
+            >>> nw.Duration() == nw.Duration("us")
+            True
+            >>> nw.Duration("us") == nw.Duration("ns")
+            False
+            >>> nw.Duration() == nw.Datetime()
+            False
+            >>> nw.Duration("ms") == nw.Duration
+            True
+        """
         if type(other) is _DurationMeta:
             return True
-        elif isinstance(other, self.__class__):
+        if isinstance(other, self.__class__):
             return self.time_unit == other.time_unit
-        else:  # pragma: no cover
-            return False
+        return False  # pragma: no cover
 
     def __hash__(self) -> int:  # pragma: no cover
         return hash((self.__class__, self.time_unit))
@@ -480,17 +599,35 @@ class Enum(DType):
 
     @property
     def categories(self) -> tuple[str, ...]:
+        """The categories in the dataset."""
         if cached := self._cached_categories:
             return cached
-        elif delayed := self._delayed_categories:
+        if delayed := self._delayed_categories:
             self._cached_categories = delayed.to_tuple()
             return self._cached_categories
-        else:  # pragma: no cover
-            msg = f"Internal structure of {type(self).__name__!r} is invalid."
-            raise TypeError(msg)
+        msg = f"Internal structure of {type(self).__name__!r} is invalid."  # pragma: no cover
+        raise TypeError(msg)  # pragma: no cover
 
-    def __eq__(self, other: object) -> bool:
-        # allow comparing object instances to class
+    def __eq__(self, other: DType | type[DType]) -> bool:  # type: ignore[override]
+        """Check if this Enum is equivalent to another DType.
+
+        Examples:
+            >>> import narwhals as nw
+            >>> nw.Enum(["a", "b", "c"]) == nw.Enum(["a", "b", "c"])
+            True
+            >>> import polars as pl
+            >>> categories = pl.Series(["a", "b", "c"])
+            >>> nw.Enum(["a", "b", "c"]) == nw.Enum(categories)
+            True
+            >>> nw.Enum(["a", "b", "c"]) == nw.Enum(["b", "a", "c"])
+            False
+            >>> nw.Enum(["a", "b", "c"]) == nw.Enum(["a"])
+            False
+            >>> nw.Enum(["a", "b", "c"]) == nw.Categorical
+            False
+            >>> nw.Enum(["a", "b", "c"]) == nw.Enum
+            True
+        """
         if type(other) is type:
             return other is Enum
         return isinstance(other, type(self)) and self.categories == other.categories
@@ -503,11 +640,11 @@ class Enum(DType):
 
 
 class Field:
-    """Definition of a single field within a `Struct` DataType.
+    """Definition of a single field within a `Struct` DType.
 
     Arguments:
         name: The name of the field within its parent `Struct`.
-        dtype: The `DataType` of the field's values.
+        dtype: The `DType` of the field's values.
 
     Examples:
        >>> import pyarrow as pa
@@ -519,14 +656,37 @@ class Field:
     """
 
     name: str
-    dtype: type[DType] | DType
+    """The name of the field within its parent `Struct`."""
+    dtype: IntoDType
+    """The `DType` of the field's values."""
 
-    def __init__(self, name: str, dtype: type[DType] | DType) -> None:
+    def __init__(self, name: str, dtype: IntoDType) -> None:
         self.name = name
         self.dtype = dtype
 
     def __eq__(self, other: Field) -> bool:  # type: ignore[override]
-        return (self.name == other.name) & (self.dtype == other.dtype)
+        """Check if this Field is equivalent to another Field.
+
+        Two fields are equivalent if they have the same name and the same dtype.
+
+        Examples:
+            >>> import narwhals as nw
+            >>> nw.Field("a", nw.String) == nw.Field("a", nw.String())
+            True
+            >>> nw.Field("a", nw.String) == nw.Field("a", nw.String)
+            True
+            >>> nw.Field("a", nw.String) == nw.Field("a", nw.Datetime)
+            False
+            >>> nw.Field("a", nw.String) == nw.Field("b", nw.String)
+            False
+            >>> nw.Field("a", nw.String) == nw.String
+            False
+        """
+        return (
+            isinstance(other, Field)
+            and (self.name == other.name)
+            and (self.dtype == other.dtype)
+        )
 
     def __hash__(self) -> int:
         return hash((self.name, self.dtype))
@@ -554,35 +714,47 @@ class Struct(NestedType):
     """
 
     fields: list[Field]
+    """The fields that make up the struct."""
 
-    def __init__(
-        self, fields: Sequence[Field] | Mapping[str, DType | type[DType]]
-    ) -> None:
+    def __init__(self, fields: Sequence[Field] | Mapping[str, IntoDType]) -> None:
         if isinstance(fields, Mapping):
             self.fields = list(starmap(Field, fields.items()))
         else:
             self.fields = list(fields)
 
     def __eq__(self, other: DType | type[DType]) -> bool:  # type: ignore[override]
-        # The comparison allows comparing objects to classes, and specific
-        # inner types to those without (eg: inner=None). if one of the
-        # arguments is not specific about its inner type we infer it
-        # as being equal. (See the List type for more info).
+        """Check if this Struct is equivalent to another DType.
+
+        Examples:
+            >>> import narwhals as nw
+            >>> nw.Struct({"a": nw.Int64}) == nw.Struct({"a": nw.Int64})
+            True
+            >>> nw.Struct({"a": nw.Int64}) == nw.Struct({"a": nw.Boolean})
+            False
+            >>> nw.Struct({"a": nw.Int64}) == nw.Struct({"b": nw.Int64})
+            False
+            >>> nw.Struct({"a": nw.Int64}) == nw.Struct([nw.Field("a", nw.Int64)])
+            True
+
+            If a parent type is not specific about its inner type, we infer it as equal
+
+            >>> nw.Struct({"a": nw.Int64}) == nw.Struct
+            True
+        """
         if type(other) is type and issubclass(other, self.__class__):
             return True
-        elif isinstance(other, self.__class__):
+        if isinstance(other, self.__class__):
             return self.fields == other.fields
-        else:
-            return False
+        return False
 
     def __hash__(self) -> int:
         return hash((self.__class__, tuple(self.fields)))
 
-    def __iter__(self) -> Iterator[tuple[str, DType | type[DType]]]:  # pragma: no cover
+    def __iter__(self) -> Iterator[tuple[str, IntoDType]]:  # pragma: no cover
         for fld in self.fields:
             yield fld.name, fld.dtype
 
-    def __reversed__(self) -> Iterator[tuple[str, DType | type[DType]]]:
+    def __reversed__(self) -> Iterator[tuple[str, IntoDType]]:
         for fld in reversed(self.fields):
             yield fld.name, fld.dtype
 
@@ -590,12 +762,8 @@ class Struct(NestedType):
         class_name = self.__class__.__name__
         return f"{class_name}({dict(self)})"
 
-    def to_schema(self) -> OrderedDict[str, DType | type[DType]]:
-        """Return Struct dtype as a schema dict.
-
-        Returns:
-            Mapping from column name to dtype.
-        """
+    def to_schema(self) -> OrderedDict[str, IntoDType]:
+        """Return Struct dtype as a schema dict."""
         return OrderedDict(self)
 
 
@@ -614,25 +782,32 @@ class List(NestedType):
        List(String)
     """
 
-    inner: DType | type[DType]
+    inner: IntoDType
+    """The DType of the values within each list."""
 
-    def __init__(self, inner: DType | type[DType]) -> None:
+    def __init__(self, inner: IntoDType) -> None:
         self.inner = inner
 
     def __eq__(self, other: DType | type[DType]) -> bool:  # type: ignore[override]
-        # This equality check allows comparison of type classes and type instances.
-        # If a parent type is not specific about its inner type, we infer it as equal:
-        # > list[i64] == list[i64] -> True
-        # > list[i64] == list[f32] -> False
-        # > list[i64] == list      -> True
+        """Check if this List is equivalent to another DType.
 
-        # allow comparing object instances to class
+        Examples:
+            >>> import narwhals as nw
+            >>> nw.List(nw.Int64) == nw.List(nw.Int64)
+            True
+            >>> nw.List(nw.Int64) == nw.List(nw.Float32)
+            False
+
+            If a parent type is not specific about its inner type, we infer it as equal
+
+            >>> nw.List(nw.Int64) == nw.List
+            True
+        """
         if type(other) is type and issubclass(other, self.__class__):
             return True
-        elif isinstance(other, self.__class__):
+        if isinstance(other, self.__class__):
             return self.inner == other.inner
-        else:
-            return False
+        return False
 
     def __hash__(self) -> int:
         return hash((self.__class__, self.inner))
@@ -657,11 +832,14 @@ class Array(NestedType):
         Array(Int32, shape=(2,))
     """
 
-    inner: DType | type[DType]
+    inner: IntoDType
+    """The DType of the values within each array."""
     size: int
+    """The size of the Array."""
     shape: tuple[int, ...]
+    """The shape of the arrays."""
 
-    def __init__(self, inner: DType | type[DType], shape: int | tuple[int, ...]) -> None:
+    def __init__(self, inner: IntoDType, shape: int | tuple[int, ...]) -> None:
         inner_shape: tuple[int, ...] = inner.shape if isinstance(inner, Array) else ()
         if isinstance(shape, int):
             self.inner = inner
@@ -681,22 +859,29 @@ class Array(NestedType):
             raise TypeError(msg)
 
     def __eq__(self, other: DType | type[DType]) -> bool:  # type: ignore[override]
-        # This equality check allows comparison of type classes and type instances.
-        # If a parent type is not specific about its inner type, we infer it as equal:
-        # > array[i64] == array[i64] -> True
-        # > array[i64] == array[f32] -> False
-        # > array[i64] == array      -> True
+        """Check if this Array is equivalent to another DType.
 
-        # allow comparing object instances to class
+        Examples:
+            >>> import narwhals as nw
+            >>> nw.Array(nw.Int64, 2) == nw.Array(nw.Int64, 2)
+            True
+            >>> nw.Array(nw.Int64, 2) == nw.Array(nw.String, 2)
+            False
+            >>> nw.Array(nw.Int64, 2) == nw.Array(nw.Int64, 4)
+            False
+
+            If a parent type is not specific about its inner type, we infer it as equal
+
+            >>> nw.Array(nw.Int64, 2) == nw.Array
+            True
+        """
         if type(other) is type and issubclass(other, self.__class__):
             return True
-        elif isinstance(other, self.__class__):
+        if isinstance(other, self.__class__):
             if self.shape != other.shape:
                 return False
-            else:
-                return self.inner == other.inner
-        else:
-            return False
+            return self.inner == other.inner
+        return False
 
     def __hash__(self) -> int:
         return hash((self.__class__, self.inner, self.shape))
@@ -746,7 +931,7 @@ class Time(TemporalType):
        Time
        >>> nw.from_native(ser_pa, series_only=True).dtype
        Time
-       >>> nw.from_native(rel).schema["t"]
+       >>> nw.from_native(rel).collect_schema()["t"]
        Time
     """
 
@@ -770,6 +955,6 @@ class Binary(DType):
         Binary
         >>> nw.from_native(ser_pa, series_only=True).dtype
         Binary
-        >>> nw.from_native(rel).schema["t"]
+        >>> nw.from_native(rel).collect_schema()["t"]
         Binary
     """
